<?php

declare(strict_types=1);

namespace Drupal\jsonapi_links\EventSubscriber;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\jsonapi_links\Exception\JsonApiLinksException;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * EventSubscriber for removing links from JSON:API responses.
 */
class ResponseSubscriber implements EventSubscriberInterface {

  /**
   * The paths to process.
   *
   * @var array
   */
  protected array $path = [];

  /**
   * A list of regexes defining patterns to ignore for removal.
   *
   * @var array
   */
  protected array $ignoreList = [];

  /**
   * The JSON:API Link module's config.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  public function __construct(
    protected ConfigFactoryInterface $configFactory,
    protected RouteMatchInterface $currentRouteMatch,
  ) {
    $this->config = $this->configFactory->get('jsonapi_links.settings');
  }

  /**
   * Checks for a JSON:API response and removes links.
   */
  public function onResponse(ResponseEvent $event): void {
    $response = $event->getResponse();
    // First, check if we have a JSON:API response.
    $header_content_type = $response->headers->get('content-type', 'unknown');
    if (!is_string($header_content_type)) {
      throw new JsonApiLinksException("Failed to get content-type header!");
    }
    if (strtolower($header_content_type) !== 'application/vnd.api+json') {
      return;
    }

    $content = $response->getContent();
    if (!is_string($content)) {
      throw new JsonApiLinksException("Failed to get content from response!");
    }
    $json = json_decode($content);
    if (!is_object($json)) {
      throw new JsonApiLinksException("Failed to decode JSON of response content!");
    }

    // Check whether we have a JSON:API JSON response.
    if (!property_exists($json, 'jsonapi')) {
      return;
    }

    // This check prevents removal of the meta.links.me info.
    $jsonapi_base_path = $this->currentRouteMatch->getParameter('jsonapi.base_path');
    if ($event->getRequest()->getPathInfo() === $jsonapi_base_path) {
      return;
    }

    // Check if we should do any processing.
    $config_remove_links = (bool) $this->config->get('remove_links');
    if ($config_remove_links === FALSE) {
      return;
    }

    $config_ignore_list = $this->config->get('ignore_list');
    // Only process the ignore list if a value has been set.
    if (is_string($config_ignore_list) && strlen($config_ignore_list) > 0) {
      $this->ignoreList = array_filter(
        array_map(
          'trim',
          explode("\n", $config_ignore_list)
        ),
        fn($value) => $value !== '',
      );
    }

    $this->removeLinks($json);

    $encoded_json = json_encode($json);
    if ($encoded_json === FALSE) {
      throw new JsonApiLinksException("Failed to encode JSON!");
    }
    $response->setContent($encoded_json);
    $event->setResponse($response);
  }

  /**
   * Unsets links in JSON:API responses.
   */
  protected function removeLinks(mixed &$data, string|int|null $node = NULL): void {
    if (!is_null($node)) {
      $this->path[] = $node;
    }

    // If the current path is on the ignore list, skip everything below that.
    if ($this->isOnIgnoreList()) {
      array_pop($this->path);
      return;
    }

    if (is_scalar($data)) {
      array_pop($this->path);
      return;
    }

    if (is_object($data)) {
      // If this is the JSON:API response, do not remove the links.
      // The links on the response may include pager links.
      if (!property_exists($data, 'jsonapi')) {
        // Meta must return an object. If links are the only property,
        // and the links are removed by this module, then an empty array
        // will be returned instead.
        // So, if meta only contains links, then remove meta as well.
        if (property_exists($data, 'meta') && property_exists($data->meta, 'links')) {
          unset($data->meta->links);
          $property_list = get_object_vars($data->meta);
          if (count($property_list) === 0) {
            unset($data->meta);
          }
        }
        if (property_exists($data, 'links')) {
          unset($data->links);
        }
      }

      foreach (get_object_vars($data) as $name => $attribute) {
        // This is safe because object vars are generated from the data object.
        $this->removeLinks($data->$name);
      }

      array_pop($this->path);
      return;
    }

    if (is_array($data)) {
      foreach ($data as $i => &$item) {
        $this->removeLinks($item, $i);
      }
    }

    array_pop($this->path);
  }

  /**
   * {@inheritDoc}
   */
  public static function getSubscribedEvents(): array {
    $events = [];
    $events[KernelEvents::RESPONSE][] = ['onResponse', 112];
    return $events;
  }

  /**
   * Checks whether a path is on the ignore list.
   */
  private function isOnIgnoreList(): bool {
    $path = implode('.', $this->path);

    for ($i = 0; $i < count($this->ignoreList); $i++) {
      if (preg_match($this->ignoreList[$i], $path) === 1) {
        return TRUE;
      }
    }

    return FALSE;
  }

}
